@@ -1,6 +1,6 @@
 /*
   ftp -- ftp protocol functions
-  Copyright (C) 1996, 1997 Dieter Baron
+  Copyright (C) 1996, 1997, 1998, 1999, 2000 Dieter Baron
 
   This file is part of cftp, a fullscreen ftp client
   The author can be contacted at <dillo@giga.or.at>
@@ -29,6 +29,7 @@
 #include <string.h>
 #include <stdlib.h>
 #include <ctype.h>
+#include <fcntl.h>
 #include <signal.h>
 #include <errno.h>
 #include <sys/types.h>
@@ -75,7 +76,14 @@ char **ftp_response;
 long ftp_response_size;
 
 FILE *conin, *conout;		/* control connection to server */
-unsigned char ftp_addr[4];	/* local ip address (for port commands) */
+
+#ifdef HAVE_STRUCT_SOCKADDR_STORAGE
+static struct sockaddr_storage _ftp_addr;
+#else
+static struct sockaddr _ftp_addr;
+#endif
+struct sockaddr *ftp_addr = (struct sockaddr *)&_ftp_addr;
+				/* local ip address (for port commands) */
 
 
 
@@ -87,11 +95,15 @@ int ftp_port(void);
 FILE *ftp_accept(int fd, char *mode);
 int ftp_mode(char m);
 int ftp_cwd(char *path);
-int ftp_cat(FILE *fin, FILE *fout, long size);
 int ftp_gethostaddr(int fd);
 void ftp_histf(char *fmt, ...);
 void ftp_hist(char *line);
 
+#ifdef ENABLE_TRANSFER_RATE
+static void _ftp_update_transfer(char *fmt, long got, long *cur,
+				 int old_sec, int new_sec);
+#endif
+
 
 
 void
@@ -114,25 +126,25 @@ ftp_init(void)
 int
 ftp_open(char *host, char *port)
 {
-	int fd;
-
-	if ((fd=sopen(host, port)) == -1)
-		return -1;
-
-	if (ftp_gethostaddr(fd) == -1) {
-		close(fd);
-		return -1;
-	}
-
-	conin = fdopen(fd, "r");
-	conout = fdopen(fd, "w");
-
-	if (!conin || !conout) {
-		close(fd);
-		return -1;
-	}
-
-	return 0;
+    int fd;
+    
+    if ((fd=sopen(host, port, AF_UNSPEC)) == -1)
+	return -1;
+    
+    if (ftp_gethostaddr(fd) == -1) {
+	close(fd);
+	return -1;
+    }
+    
+    conin = fdopen(fd, "r");
+    conout = fdopen(fd, "w");
+    
+    if (!conin || !conout) {
+	close(fd);
+	return -1;
+    }
+    
+    return 0;
 }
 
 
@@ -175,6 +187,15 @@ ftp_login(char *host, char *user, char *pass)
 
 
 
+int
+ftp_site(char *cmd)
+{
+    ftp_put("%s", cmd);
+    return ftp_resp();
+}
+
+
+
 int
 ftp_reconnect(void)
 {
@@ -218,6 +239,7 @@ ftp_reconnect(void)
     status.remote.path = strdup(ftp_lcwd);
     status_do(bs_remote);
     ftp_pcwd = NULL;
+    ftp_curmode = ' ';
 
     return 0;
 }
@@ -227,20 +249,20 @@ ftp_reconnect(void)
 int
 ftp_close(void)
 {
-	int err = 0;
-
-	if (conin == NULL)
-	    return 0;
-
-	ftp_put("quit");
-	if (ftp_resp() != 221)
-		err = 1;
-
-	fclose(conin);
-	fclose(conout);
-	conin = conout = NULL;
-
-	return err;
+    int err = 0;
+    
+    if (conin == NULL)
+	return 0;
+    
+    ftp_put("quit");
+    if (ftp_resp() != 221)
+	err = 1;
+    
+    fclose(conin);
+    fclose(conout);
+    conin = conout = NULL;
+    
+    return err;
 }
 
 
@@ -248,47 +270,47 @@ ftp_close(void)
 directory *
 ftp_list(char *path)
 {
-	directory *dir;
-	int fd;
-	FILE *f;
-
-	if (ftp_mode('a') == -1 || ftp_cwd(path) == -1)
-		return NULL;
-
-	free(status.remote.path);
-	status.remote.path = strdup(path);
-	status_do(bs_remote);
-
-	if ((fd=ftp_port()) == -1)
-		return NULL;
-	
-	ftp_put("list");
-	if (ftp_resp() != 150) {
-		close(fd);
-		dir = (directory *)malloc(sizeof(directory));
-		dir->line = (direntry *)malloc(sizeof(direntry));
-		dir->path = strdup(path);
-		dir->len = 0;
-		dir->cur = dir->top = 0;
-		dir->size = sizeof(struct direntry);
-		dir->line->line = strdup("");
-		dir->line->type = 'x';
-		dir->line->name = strdup("");
-		dir->line->link = NULL;
-		return dir;
-	}
-	if ((f=ftp_accept(fd, "r")) == NULL)
-	    return NULL;
-	
-	dir = read_dir(f);
-	if (dir)
-	    dir->path = strdup(path);
-
-	fclose(f);
-
-	ftp_resp();
-
+    directory *dir;
+    int fd;
+    FILE *f;
+    
+    if (ftp_mode('a') == -1 || ftp_cwd(path) == -1)
+	return NULL;
+    
+    free(status.remote.path);
+    status.remote.path = strdup(path);
+    status_do(bs_remote);
+    
+    if ((fd=ftp_port()) == -1)
+	return NULL;
+    
+    ftp_put("list");
+    if (ftp_resp() != 150) {
+	close(fd);
+	dir = (directory *)malloc(sizeof(directory));
+	dir->line = (direntry *)malloc(sizeof(direntry));
+	dir->path = strdup(path);
+	dir->len = 0;
+	dir->cur = dir->top = 0;
+	dir->size = sizeof(struct direntry);
+	dir->line->line = strdup("");
+	dir->line->type = 'x';
+	dir->line->name = strdup("");
+	dir->line->link = NULL;
 	return dir;
+    }
+    if ((f=ftp_accept(fd, "r")) == NULL)
+	return NULL;
+    
+    dir = read_dir(f);
+    if (dir)
+	dir->path = strdup(path);
+    
+    fclose(f);
+    
+    ftp_resp();
+    
+    return dir;
 }
 
 
@@ -296,56 +318,69 @@ ftp_list(char *path)
 directory *
 ftp_cd(char *wd, int force)
 {
-	directory *dir;
-	char *nwd;
+    directory *dir;
+    char *nwd;
+    
+    nwd = canonical(wd, NULL);
+    
+    dir = get_dir(nwd, force);
+    if (dir != NULL) {
+	free(ftp_lcwd);
+	ftp_lcwd = nwd;
 	
-	nwd = canonical(wd, NULL);
-
-	dir = get_dir(nwd, force);
-	if (dir != NULL) {
-		free(ftp_lcwd);
-		ftp_lcwd = nwd;
-		
-		free(status.remote.path);
-		status.remote.path = strdup(ftp_lcwd);
-		status_do(bs_remote);
-	}
-	else
-		free(nwd);
-
-	return dir;
+	free(status.remote.path);
+	status.remote.path = strdup(ftp_lcwd);
+	status_do(bs_remote);
+    }
+    else
+	free(nwd);
+    
+    return dir;
 }
 	
 
 
 FILE *
-ftp_retr(char *file, int mode)
+ftp_retr(char *file, int mode, long *startp, long *sizep)
 {
-	int fd;
-	char *dir, *name, *can;
-	FILE *fin;
-
-	can = canonical(file, NULL);
-	dir = dirname(can);
-	name = (char *)basename(can);
-	
-	if (ftp_mode(mode) == -1 || ftp_cwd(dir) == -1)
-		return NULL;
-
-	if ((fd=ftp_port()) == -1)
-		return NULL;
-	
-	ftp_put("retr %s", name);
-	if (ftp_resp() != 150) {
-		close(fd);
-		return NULL;
+    int fd;
+    char *dir, *name, *can, *p;
+    FILE *fin;
+    
+    can = canonical(file, NULL);
+    dir = dirname(can);
+    name = (char *)basename(can);
+    
+    if (ftp_mode(mode) == -1 || ftp_cwd(dir) == -1)
+	return NULL;
+    
+    if ((fd=ftp_port()) == -1)
+	return NULL;
+    
+    if (startp && *startp > 0) {
+	ftp_put("rest %ld", *startp);
+	if (ftp_resp() != 350)
+	    *startp = 0;
+    }
+    
+    ftp_put("retr %s", name);
+    if (ftp_resp() != 150) {
+	close(fd);
+	return NULL;
+    }
+    if (sizep != NULL) {
+	/* XXX: check how other servers format 150s */
+	if (strcmp(ftp_last_resp+strlen(ftp_last_resp)-8, " bytes).") == 0) {
+	    if ((p=strrchr(ftp_last_resp, '(')) != NULL)
+		*sizep = strtol(p+1, NULL, 10);
 	}
-	if ((fin=ftp_accept(fd, "r")) == NULL) {
-		close(fd);
-		return NULL;
-	}
-
-	return fin;
+    }
+    if ((fin=ftp_accept(fd, "r")) == NULL) {
+	close(fd);
+	return NULL;
+    }
+    
+    return fin;
 }
 
 
@@ -353,31 +388,31 @@ ftp_retr(char *file, int mode)
 FILE *
 ftp_stor(char *file, int mode)
 {
-	int fd;
-	char *dir, *name, *can;
-	FILE *fin;
-
-	can = canonical(file, NULL);
-	dir = dirname(can);
-	name = (char *)basename(can);
-	
-	if (ftp_mode(mode) == -1 || ftp_cwd(dir) == -1)
-		return NULL;
-
-	if ((fd=ftp_port()) == -1)
-		return NULL;
-	
-	ftp_put("stor %s", name);
-	if (ftp_resp() != 150) {
-		close(fd);
-		return NULL;
-	}
-	if ((fin=ftp_accept(fd, "w")) == NULL) {
-		close(fd);
-		return NULL;
-	}
-
-	return fin;
+    int fd, resp;
+    char *dir, *name, *can;
+    FILE *fin;
+    
+    can = canonical(file, NULL);
+    dir = dirname(can);
+    name = (char *)basename(can);
+    
+    if (ftp_mode(mode) == -1 || ftp_cwd(dir) == -1)
+	return NULL;
+    
+    if ((fd=ftp_port()) == -1)
+	return NULL;
+    
+    ftp_put("stor %s", name);
+    if ((resp=ftp_resp()) != 150 && resp != 125) {
+	close(fd);
+	return NULL;
+    }
+    if ((fin=ftp_accept(fd, "w")) == NULL) {
+	close(fd);
+	return NULL;
+    }
+    
+    return fin;
 }
 
 
@@ -397,6 +432,18 @@ ftp_fclose(FILE *f)
 
 
 
+int
+ftp_mkdir(char *path)
+{
+    ftp_put("mkd %s", path);
+    if (ftp_resp() != 257)
+	return -1;
+    
+    return 0;
+}
+
+
+
 int
 ftp_noop(void)
 {
@@ -430,19 +477,23 @@ ftp_pwd(void)
     dir[e-s-1] = '\0';
 
     if (ftp_dosnames == -1) {
-	if (isalpha(dir[0]) && dir[1] == ':' && dir[2] == '\\')
+	if ((isalpha(dir[0]) && dir[1] == ':' && dir[2] == '\\'))
 	    ftp_dosnames = 1;
+	else if (dir[0] == '\\')
+	    ftp_dosnames = 2;
 	else
 	    ftp_dosnames = 0;
     }
 
-    if (ftp_dosnames == 1) {
+    if (ftp_dosnames == 1 || ftp_dosnames == 2) {
+	if (ftp_dosnames == 1) {
 	    strncpy(dir+1, s+1, e-s-1);
 	    dir[e-s] = '\0';
 	    dir[0] = '/';
-	    for (s=dir; *s; s++)
-		if (*s == '\\')
-		    *s = '/';
+	}
+	for (s=dir; *s; s++)
+	    if (*s == '\\')
+		*s = '/';
     }
 
     free(ftp_pcwd);
@@ -456,33 +507,33 @@ ftp_pwd(void)
 char *
 ftp_gets(FILE *f)
 {
-	char buf[8192], *line;
-	int l;
-
-	if (fgets(buf, 8192, f) == NULL)
-		return NULL;
-
-	line = strdup(buf);
-	l = strlen(line);
-
-	while (line[l-1] != '\n') {
-		if (fgets(buf, 8192, f) == NULL) {
-			free(line);
-			return NULL;
-		}
-		l += strlen(buf);
-		if ((line=realloc(line, l+1)) == NULL) {
-		    disp_status("malloc failure");
-		    return NULL;
-		}
-		strcpy(line+l, buf);
+    char buf[8192], *line;
+    int l;
+    
+    if (fgets(buf, 8192, f) == NULL)
+	return NULL;
+    
+    line = strdup(buf);
+    l = strlen(line);
+    
+    while (line[l-1] != '\n') {
+	if (fgets(buf, 8192, f) == NULL) {
+	    free(line);
+	    return NULL;
 	}
-	if (line[l-2] == '\r')
-		line[l-2] = '\0';
-	else
-		line[l-1] = '\0';
-
-	return line;
+	l += strlen(buf);
+	if ((line=realloc(line, l+1)) == NULL) {
+	    disp_status("malloc failure");
+	    return NULL;
+	}
+	strcpy(line+l, buf);
+    }
+    if (line[l-2] == '\r')
+	line[l-2] = '\0';
+    else
+	line[l-1] = '\0';
+    
+    return line;
 }
 
 
@@ -490,34 +541,34 @@ ftp_gets(FILE *f)
 int
 ftp_put(char *fmt, ...)
 {
-	char buf[8192];
-	va_list argp;
-
-	if (conout == NULL) {
-	    disp_status("not connected");
-	    return -1;
-	}
-
-	va_start(argp, fmt);
-	vsprintf(buf, fmt, argp);
-	va_end(argp);
-
-	if (strncmp(buf, "pass ", 5) == 0 && ftp_anon == 0) {
-		disp_status("-> pass ********");
-		ftp_hist(strdup("-> pass ********"));
-	}
-	else {
-		disp_status("-> %s", buf);
-		ftp_histf("-> %s", buf);
-	}
-	fprintf(conout, "%s\r\n", buf);
-
-	if (fflush(conout) || ferror(conout)) {
-	    disp_status("error writing to server: %s", strerror(errno));
-	    return -1;
-	}
-
-	return 0;
+    char buf[8192];
+    va_list argp;
+    
+    if (conout == NULL) {
+	disp_status("not connected");
+	return -1;
+    }
+    
+    va_start(argp, fmt);
+    vsprintf(buf, fmt, argp);
+    va_end(argp);
+    
+    if (strncmp(buf, "pass ", 5) == 0 && ftp_anon == 0) {
+	disp_status("-> pass ********");
+	ftp_hist(strdup("-> pass ********"));
+    }
+    else {
+	disp_status("-> %s", buf);
+	ftp_histf("-> %s", buf);
+    }
+    fprintf(conout, "%s\r\n", buf);
+    
+    if (fflush(conout) || ferror(conout)) {
+	disp_status("error writing to server: %s", strerror(errno));
+	return -1;
+    }
+    
+    return 0;
 }	
 
 
@@ -644,21 +695,109 @@ void ftp_unresp(int resp)
 int
 ftp_port(void)
 {
-	unsigned long host;
-	int fd, port;
+    int fd, port, val, i, delim;
+    int len;
+    unsigned int iaddr;
+    unsigned char addr[4];
+    char baddr[16], bport[6], *s, *e;
+    char *saddr, *sport;
+#ifdef HAVE_STRUCT_SOCKADDR_STORAGE
+    struct sockaddr_storage sa;
+#else
+    struct sockaddr sa;
+#endif
+    struct sockaddr_in *sin;
 
-	if ((fd=spassive(&host, &port)) == -1)
-		return -1;
+    if (!opt_pasv) {
+	sin = (struct sockaddr_in *)ftp_addr;
 
-	ftp_put("port %d,%d,%d,%d,%d,%d", (int)ftp_addr[0],
-		(int)ftp_addr[1], (int)ftp_addr[2], (int)ftp_addr[3], 
-		port>>8, port&0xff);
-	if (ftp_resp() != 200) {
-		close(fd);
-		return -1;
+	len = sizeof(sa);
+	if ((fd=spassive(ftp_addr->sa_family,
+			 (struct sockaddr *)&sa, &len)) == -1)
+	    return -1;
+
+	port = htons(((struct sockaddr_in *)&sa)->sin_port);
+
+	if (ftp_addr->sa_family == AF_INET) {
+	    iaddr = htonl(sin->sin_addr.s_addr);
+	    ftp_put("port %d,%d,%d,%d,%d,%d",
+		    (iaddr>>24) & 0xff, (iaddr>>16) & 0xff,
+		    (iaddr>>8) & 0xff, iaddr & 0xff,
+		    (port>>8) & 0xff, port & 0xff);
 	}
+	else {
+	    ftp_put("eprt |%d|%s|%d|",
+		    /* proto no */ 2,
+		    sockaddr_ntop(ftp_addr), port);
+	}
+	
+	if (ftp_resp() != 200) {
+	    close(fd);
+	    return -1;
+	}
+    }
+    else {
+	if (ftp_addr->sa_family == AF_INET) {
+	    ftp_put("pasv");
+	    if (ftp_resp() != 227)
+		return -1;
 
-	return fd;
+	    if ((s=strchr(ftp_last_resp, ',')) == NULL)
+		return -1;
+	    
+	    while (isdigit(*(--s)))
+		;
+	    s++;
+	    
+	    for (i=0; i<4; i++) {
+		val = strtol(s, &e, 10);
+		if (val < 0 || val > 255 || *e != ',')
+		    return -1;
+		addr[i] = val;
+		s = e+1;
+	    }
+	    port = strtol(s, &e, 10);
+	    if (port < 0 || port > 255 || *e != ',')
+		return -1;
+	    s = e+1;
+	    val = strtol(s, &e, 10);
+	    if (val < 0 || val > 255)
+		return -1;
+	    port = port*256+val;
+	    
+	    sprintf(baddr, "%d.%d.%d.%d",
+		    addr[0], addr[1], addr[2], addr[3]);
+	    sprintf(bport, "%d", port);
+	    saddr = baddr;
+	    sport = bport;
+	}
+	else { /* inet 6 passive */
+	    ftp_put("epsv");
+	    if (ftp_resp() != 229)
+		return -1;
+
+	    if ((s=strchr(ftp_last_resp, '(')) == NULL)
+		return -1;
+	    delim = s[1];
+
+	    if ((e=strchr(s+2, delim)) == NULL
+		|| (e=strchr(e+1, delim)) == NULL)
+		return -1;
+
+	    s = e+1;
+	    if ((e=strchr(s, delim)) == NULL)
+		return -1;
+	    if (e-s>5)
+		return -1;
+	    strncpy(bport, s, e-s);
+	    bport[e-s] = '\0';
+	    sport = bport;
+	    saddr = ftp_host;
+	}
+	fd = sopen(saddr, sport, ftp_addr->sa_family);
+    }
+    
+    return fd;
 }
 
 
@@ -666,16 +805,21 @@ ftp_port(void)
 FILE *
 ftp_accept(int fd, char *mode)
 {
-	int len, ns;
-	struct sockaddr addr;
+    int len, ns;
+    struct sockaddr addr;
 
+    if (!opt_pasv) {
 	len = sizeof(addr);
 	if ((ns=accept(fd, &addr, &len)) == -1)
-		return NULL;
-
-	close(fd);
+	    return NULL;
 	
-	return fdopen(ns, mode);
+	close(fd);
+	fd = ns;
+    }
+	
+    fcntl(fd, F_SETFD, 1); /* XXX: error check */
+    
+    return fdopen(fd, mode);
 }
 
 
@@ -683,15 +827,15 @@ ftp_accept(int fd, char *mode)
 int
 ftp_mode(char m)
 {
-	if (m == ftp_curmode)
-		return 0;
-
-	ftp_put("type %c", toupper(m));
-	if (ftp_resp() != 200)
-		return -1;
-
-	ftp_curmode = m;
+    if (m == ftp_curmode)
 	return 0;
+    
+    ftp_put("type %c", toupper(m));
+    if (ftp_resp() != 200)
+	return -1;
+    
+    ftp_curmode = m;
+    return 0;
 }
 
 
@@ -700,16 +844,18 @@ int
 ftp_cwd(char *path)
 {
     char *s, *e;
+    int off;
     
     if (ftp_pcwd && strcmp(path, ftp_pcwd) == 0)
 	return 0;
 
-    if (ftp_dosnames == 1) {
-	for (s=path+1; *s; s++)
+    if (ftp_dosnames == 1 || ftp_dosnames == 2) {
+	off = (ftp_dosnames == 1) ? 1 : 0;
+	for (s=path+off; *s; s++)
 	    if (*s == '/')
 		*s = '\\';
-	ftp_put("cwd %s", path+1);
-	for (s=path+1; *s; s++)
+	ftp_put("cwd %s", path+off);
+	for (s=path+off; *s; s++)
 	    if (*s == '\\')
 		*s = '/';
     }
@@ -723,6 +869,8 @@ ftp_cwd(char *path)
 	    && (e=strchr(s+1, '"')) != NULL) {
 	    if (isalpha(s[1]) && s[2] == ':' && s[3] == '\\')
 		ftp_dosnames = 1;
+	    else if (s[1] == '\\')
+		ftp_dosnames = 2;
 	    else
 		ftp_dosnames = 0;
 	}
@@ -739,22 +887,49 @@ ftp_cwd(char *path)
 
 
 int
-ftp_cat(FILE *fin, FILE *fout, long size)
+ftp_cat(FILE *fin, FILE *fout, long start, long size)
 {
-    time_t oldt, newt;
     char buf[4096], fmt[4096];
     int c, n, err;
-    long got = 0;
+    long got;
+#ifdef ENABLE_TRANSFER_RATE
+    struct itimerval itv;
+    long cur[4];
+    int old_alarm;
+#else
+    time_t oldt, newt;
+#endif
 
-    if (size >= 0)
+    if (size >= 0) {
+#ifdef ENABLE_TRANSFER_RATE
+	sprintf(fmt, "transferred %%ld/%ld "
+		"(total: %%.2fkb/s, current: %%.2fkb/s)",
+		size);
+#else
 	sprintf(fmt, "transferred %%ld/%ld", size);
-    else
+#endif
+    }
+    else {
+#ifdef ENABLE_TRANSFER_RATE
+	strcpy(fmt, "transferred %ld (total: %%.2fkb/s, current: %%.2fkb/s)");
+#else
 	strcpy(fmt, "transferred %ld");
-	
-    oldt = 0;
+#endif
+    }
 
+    got = start;
     signal(SIGINT, sig_remember);
 
+#ifdef ENABLE_TRANSFER_RATE
+    cur[0] = cur[1] = cur[2] = cur[3] = 0;
+    old_alarm = sig_alarm = sig_intr = 0;
+    itv.it_value.tv_sec = itv.it_interval.tv_sec = 1;
+    itv.it_value.tv_usec = itv.it_interval.tv_usec = 0;
+    setitimer(ITIMER_REAL, &itv, NULL);
+#else
+    oldt = 0;
+#endif
+
     if (ftp_curmode == 'i')
 	while ((n=fread(buf, 1, 4096, fin)) > 0) {
 	    if (sig_intr)
@@ -763,10 +938,18 @@ ftp_cat(FILE *fin, FILE *fout, long size)
 		break;
 	    }
 	    got += n;
+#ifdef ENABLE_TRANSFER_RATE
+	    if (old_alarm != sig_alarm) {
+		_ftp_update_transfer(fmt, got-start, cur,
+				     old_alarm, sig_alarm);
+		old_alarm = sig_alarm;
+	    }
+#else
 	    if ((newt=time(NULL)) != oldt) {
 		disp_status(fmt, got);
 		oldt = newt;
 	    }
+#endif
 	}
     else
 	while ((c=getc(fin)) != EOF) {
@@ -784,14 +967,26 @@ ftp_cat(FILE *fin, FILE *fout, long size)
 	    if (ferror(fout))
 		break;
 
+#ifdef ENABLE_TRANSFER_RATE
+	    if (old_alarm != sig_alarm) {
+		_ftp_update_transfer(fmt, got-start, cur,
+				     old_alarm, sig_alarm);
+		old_alarm = sig_alarm;
+	    }
+#else
 	    if (got%512 == 0 && (newt=time(NULL)) != oldt) {
 		disp_status(fmt, got);
 		oldt = newt;
-	    }
-
+	    } 
+#endif
 	}
 
     signal(SIGINT, sig_end);
+#ifdef ENABLE_TRANSFER_RATE
+    itv.it_value.tv_sec = itv.it_interval.tv_sec = 0;
+    itv.it_value.tv_usec = itv.it_interval.tv_usec = 0;
+    setitimer(ITIMER_REAL, &itv, NULL);
+#endif
 
     if (ferror(fin) || sig_intr) {
 	errno = 0;
@@ -812,27 +1007,65 @@ ftp_cat(FILE *fin, FILE *fout, long size)
 
 
 
+#ifdef ENABLE_TRANSFER_RATE
+static void
+_ftp_update_transfer(char *fmt, long got, long *cur, int old_sec, int new_sec)
+{
+    long step;
+    int nsec;
+
+    nsec = new_sec-old_sec;
+
+    switch (nsec) {
+    case 0:
+	break;
+
+    case 1:
+	cur[0] = cur[1];
+	cur[1] = cur[2];
+	cur[2] = cur[3];
+	break;
+
+    case 2:
+	cur[0] = cur[2];
+	cur[1] = cur[3];
+	cur[2] = cur[1] + (got-cur[1])/2;
+	break;
+
+    default:
+	step = (got-cur[3]) / nsec;
+	cur[0] = cur[3]+step*(nsec-3);
+	cur[1] = cur[3]+step*(nsec-2);
+	cur[2] = cur[3]+step*(nsec-1);
+	break;
+    }
+    cur[3] = got;
+
+    disp_status(fmt, got, got/(float)(new_sec*1024),
+		(cur[3]-cur[0])/((float)3*1024));
+}
+#endif /* ENABLE_TRANSFER_RATE */
+
+
+
 int
 ftp_gethostaddr(int fd)
 {
-	struct sockaddr_in addr;
-	int len;
+    int len;
 	
-	len = sizeof(addr);
-	if (getsockname(fd, (struct sockaddr* )&addr, &len) == -1) {
-	    if (disp_active)
-		disp_status("can't get host address: %s",
-			    strerror(errno));
-	    else
-		fprintf(stderr, "%s: can't get host addres: %s\n",
-			prg, strerror(errno));
-	    
-	    return -1;
-	}
-
-	memcpy(ftp_addr, (char *)&addr.sin_addr.s_addr, 4);
-
-	return 0;
+    len = sizeof(_ftp_addr);
+    if (getsockname(fd, ftp_addr, &len) == -1) {
+	if (disp_active)
+	    disp_status("can't get host address: %s",
+			strerror(errno));
+	else
+	    fprintf(stderr, "%s: can't get host addres: %s\n",
+		    prg, strerror(errno));
+	
+	return -1;
+    }
+    
+    return 0;
 }
 
 
