@@ -28,11 +28,11 @@ struct uoption {
 extern struct uoption option[];
 
 extern int opt_mode;
+extern int opt_pasv;
 extern int opt_tagchar;
 extern int ftp_hist_size;
 extern int opt_beep;
 extern int opt_scrlimit;
-extern int opt_emacs_status;
 extern char *opt_pager;
 extern int opt_sort;
 extern int opt_wrap;
