@@ -3,7 +3,7 @@
 
 /*
   signals.h -- signal handling functions
-  Copyright (C) 1996, 1997 Dieter Baron
+  Copyright (C) 1996, 1997, 1998, 1999, 2000 Dieter Baron
 
   This file is part of cftp, a fullscreen ftp client
   The author can be contacted at <dillo@giga.or.at>
@@ -25,7 +25,7 @@
 
 
 
-volatile int sig_intr, sig_pipe;
+extern volatile int sig_intr, sig_pipe, sig_alarm;
 
 int signals_init(void);
 void sig_end(int sno);
