@@ -3,7 +3,7 @@
 
 /*
   tag.h -- tagging
-  Copyright (C) 1996, 1997 Dieter Baron
+  Copyright (C) 1996, 1997, 1998, 1999, 2000 Dieter Baron
 
   This file is part of cftp, a fullscreen ftp client
   The author can be contacted at <dillo@giga.or.at>
@@ -25,6 +25,8 @@
 
 
 
+#include "directory.h"
+
 struct tagentry {
     char *line, *name;
     struct tagentry *next, *prev;
@@ -53,4 +55,6 @@ void tag_delete(int n);
 void tag_clear(void);
 int tag_anytags(void);
 
+void change_curdir(directory *dir);
+
 #endif /* tag.h */
