@@ -3,7 +3,7 @@
 
 /*
   util.h -- auxiliary functions
-  Copyright (C) 1996, 1997 Dieter Baron
+  Copyright (C) 1996, 1997, 1998, 1999, 2000 Dieter Baron
 
   This file is part of cftp, a fullscreen ftp client
   The author can be contacted at <dillo@giga.or.at>
@@ -28,6 +28,8 @@
 char *canonical(char *path, char *current);
 char *dirname(char *name);
 char *local_exp(char *path);
-char *argstostr(char **args);
+int parse_url(char *url, char **user, char **pass,
+	      char **host, char **port, char **dir);
+char *args_to_string(char **args);
 
 #endif /* util.h */
